INIT = "__init__"
NEW = "__new__"
CLASS = "__class__"
GENERIC_CONSTRUCTOR = "__class_getitem__"
SPECIAL_CONSTRUCTOR = "__getitem__"
ORIGIN = "__origin__"
ORIGINAL_CLASS = "__orig_class__"
ARGS = "__args__"
PARAMETERS = "__parameters__"
BASES = "__bases__"
NAME = "__name__"
QUALIFIED_NAME = "__qualname__"
MODULE = "__module__"
ANNOTATIONS = "__annotations__"
DICT = "__dict__"
SELF = "__self__"
GLOBALS = "__globals__"
IM_CLASS = "im_class"
WRAPPED = "__wrapped__"
BOUND = "__bound__"
CONSTRAINTS = "__constraints__"
TYPE_PARAMS = "__type_params__"
DEFAULT = "__default__"